// hlavickove subory *******************************************************
#include "../includes.h"
// vlastny .h subor
#include "loadmd3.h"

// konstruktor *************************************************************
CLoadMD3::CLoadMD3(void)
{
 fp=NULL; // subor

 frames=NULL;   // snimky
 tags=NULL;     // znacky
 surfaces=NULL; // sietove modely
 links=NULL;    // prepojenia

 // dekodovanie normal
 float lat,lng;
 // priprava tabulky s predpripravenymi normalami
 loopj(256) {
  loopi(256) {
   lat=i*2.0f*M_PI/255.0f;
   lng=j*2.0f*M_PI/255.0f;

   nxyz[i][j][0]=(float) (cos(lat)*sin(lng));
   nxyz[i][j][1]=(float) (sin(lat)*sin(lng));
   nxyz[i][j][2]=(float) cos(lng);
  }
 }

 actframe=0; nextframe=0;
 ipol=0.0f;
 stopaction=0;

 actV=0; ofsV=0; ipos=0;

 // prvy, posledny, aktualny odtlacok (casti odtlacku),
 // oznacenie casti posledneho pridaneho odtlacku
 firstD=NULL; lastD=NULL; curD=NULL; cD=NULL; lastG=NULL;
}

// destruktor **************************************************************
CLoadMD3::~CLoadMD3(void)
{
 DELETEA(frames);      // snimky
 DELETEA(tags);        // znacky
 if (surfaces!=NULL) { // sietove modely
  loopk(header.numSurfaces) {
   DELETEA(surfaces[k].shaders); DELETEA(surfaces[k].triangles);
   DELETEA(surfaces[k].texcoords); DELETEA(surfaces[k].vertices);

   DELETEA(surfaces[k].tris_v); // pole trojuholnikov
   DELETEA(surfaces[k].tris_t); // pole texturovacich suradnic
  }
 }
 DELETEA(surfaces);
 DELETEA(links);       // prepojenia

 // zmazanie vsetkych odtlackov (vsetkych casti)
 delAllDecals();
}

// nacitanie obsahu MD3 suboru *********************************************
int CLoadMD3::load(const char *FNAME)
{
 // otvorenie MD3 suboru
 fp=fopen(FNAME,"rb");
 if (fp==NULL) { printf("Nemozno otvorit MD3 subor: %s!\n",FNAME); return(1); }

 long cur_offset=ftell(fp); // aktualna pozicia v subore

 // nacitanie hlavicky MD3 suboru
 fr=fread(&header,sizeof(MD3header),1,fp);
 // test verzie MD3 suboru
 if (header.version!=15) {
  printf("Nepodporovana verzia MD3 suboru: %i!\n",header.version); fclose(fp); return(1);
 }
 // INFO: MD3 subor
 printf("MD3 subor: %s\n",FNAME);
 // INFO: hlavicka MD3 suboru
 printf("ident: %c%c%c%c ",header.ident[0],header.ident[1],header.ident[2],header.ident[3]);
 printf("name: %s ",header.name);
 printf("frames: %i ",header.numFrames);
 printf("tags: %i ",header.numTags);
 printf("surfaces: %i ",header.numSurfaces);
 printf("skins: %i\n",header.numSkins);

 // vyhradenie pamate
 frames=(MD3frame *) malloc(sizeof(MD3frame)*header.numFrames);
 tags=(MD3tag *) malloc(sizeof(MD3tag)*header.numFrames*header.numTags);
 surfaces=(MD3surface *) malloc(sizeof(MD3surface)*header.numSurfaces);
 links=(CLoadMD3 **) malloc(header.numTags);
 if ((frames==NULL)||(tags==NULL)||(surfaces==NULL)||(links==NULL)) {
  printf("Nemozno vyhradit potrebnu pamat!\n"); fclose(fp); return(1);
 }

 // nacitanie snimok
 fseek(fp,cur_offset+header.ofsFrames,SEEK_SET);
 fr=fread(frames,sizeof(MD3frame),header.numFrames,fp);

 // nacitanie znaciek
 fseek(fp,cur_offset+header.ofsTags,SEEK_SET);
 fr=fread(tags,sizeof(MD3tag),header.numFrames*header.numTags,fp);
 // INFO: znacky MD3 suboru
 if (header.numTags!=0) {
  printf("znacky: "); loopi(header.numTags) printf("%i %s ",i,tags[i].name); printf("\n");
 }

 // nacitanie sietovych modelov
 cur_offset+=header.ofsSurfaces;
 loopi(header.numSurfaces) { // pre vsetky sietove modely
  fseek(fp,cur_offset,SEEK_SET);
  // hlavicka sietoveho modelu
  fr=fread(&surfaces[i].header,sizeof(MD3surface_header),1,fp);
  // vyhradenie pamate
  surfaces[i].shaders=(MD3shader *) malloc(sizeof(MD3shader)*surfaces[i].header.numShaders);
  surfaces[i].triangles=(MD3triangle *) malloc(sizeof(MD3triangle)*surfaces[i].header.numTriangles);
  surfaces[i].texcoords=(MD3st *) malloc(sizeof(MD3st)*surfaces[i].header.numVerts);
  surfaces[i].vertices=(MD3vertex *) malloc(sizeof(MD3vertex)*surfaces[i].header.numFrames*surfaces[i].header.numVerts);
  if ((surfaces[i].shaders==NULL)||(surfaces[i].triangles==NULL)||
      (surfaces[i].texcoords==NULL)||(surfaces[i].vertices==NULL)) {
   printf("Nemozno vyhradit potrebnu pamat!\n"); fclose(fp); return(1);
  }
  // shadery
  fseek(fp,cur_offset+surfaces[i].header.ofsShaders,SEEK_SET);
  fr=fread(surfaces[i].shaders,sizeof(MD3shader),surfaces[i].header.numShaders,fp);
  // odkazy na vrcholy trojuholnikov
  fseek(fp,cur_offset+surfaces[i].header.ofsTriangles,SEEK_SET);
  fr=fread(surfaces[i].triangles,sizeof(MD3triangle),surfaces[i].header.numTriangles,fp);
  // texturovacie suradnice
  fseek(fp,cur_offset+surfaces[i].header.ofsSt,SEEK_SET);
  fr=fread(surfaces[i].texcoords,sizeof(MD3st),surfaces[i].header.numVerts,fp);
  // vrcholy
  fseek(fp,cur_offset+surfaces[i].header.ofsXyznormal,SEEK_SET);
  fr=fread(surfaces[i].vertices,sizeof(MD3vertex),surfaces[i].header.numFrames*surfaces[i].header.numVerts,fp);
  // presun na dalsi sietovy model
  cur_offset+=surfaces[i].header.ofsEnd;

  // presun trojuholnikov a texturovacich suradnic do poli
  surfaces[i].tris_v=(vec3 *) malloc(3*sizeof(vec3)*surfaces[i].header.numFrames*surfaces[i].header.numTriangles);
  surfaces[i].tris_t=(vec2 *) malloc(3*sizeof(vec2)*surfaces[i].header.numTriangles);
  if ((surfaces[i].tris_v==NULL)||(surfaces[i].tris_t==NULL)) {
   printf("Nemozno vyhradit potrebnu pamat!\n"); return(1);
  }
  printf("surface: %i, triangles: %i\n",i,surfaces[i].header.numTriangles);

  ipos=0; // pozicia v poli trojuholnikov
  loopk(surfaces[i].header.numFrames) { // pre vsetky snimky
   // 1. vrchol zo sady vrcholov konkretneho sietoveho modelu patriacich snimke k
   ofsV=k*surfaces[i].header.numVerts;

   loopj(surfaces[i].header.numTriangles) { // pre vsetky trojuholniky
    actV=ofsV+surfaces[i].triangles[j].indexes[0]; // odkaz na 1. vrchol trojuholnika
    surfaces[i].tris_v[ipos+0].set((float) (surfaces[i].vertices[actV].v[0])/2000.0f,
                                   (float) (surfaces[i].vertices[actV].v[1])/2000.0f,
                                   (float) (surfaces[i].vertices[actV].v[2])/2000.0f);
    actV=ofsV+surfaces[i].triangles[j].indexes[1]; // odkaz na 2. vrchol trojuholnika
    surfaces[i].tris_v[ipos+2].set((float) (surfaces[i].vertices[actV].v[0])/2000.0f,
                                   (float) (surfaces[i].vertices[actV].v[1])/2000.0f,
                                   (float) (surfaces[i].vertices[actV].v[2])/2000.0f);
    actV=ofsV+surfaces[i].triangles[j].indexes[2]; // odkaz na 3. vrchol trojuholnika
    surfaces[i].tris_v[ipos+1].set((float) (surfaces[i].vertices[actV].v[0])/2000.0f,
                                   (float) (surfaces[i].vertices[actV].v[1])/2000.0f,
                                   (float) (surfaces[i].vertices[actV].v[2])/2000.0f);
    ipos+=3; // posun v poli trojuholnikov
   } // pre vsetky trojuholniky
  } // pre vsetky snimky

  ipos=0; // pozicia v poli texturovacich suradnic
  loopj(surfaces[i].header.numTriangles) { // pre vsetky trojuholniky
   actV=surfaces[i].triangles[j].indexes[0]; // odkaz na 1. vrchol trojuholnika
   surfaces[i].tris_t[ipos+0].set(surfaces[i].texcoords[actV].st[0],surfaces[i].texcoords[actV].st[1]);
   actV=surfaces[i].triangles[j].indexes[1]; // odkaz na 2. vrchol trojuholnika
   surfaces[i].tris_t[ipos+2].set(surfaces[i].texcoords[actV].st[0],surfaces[i].texcoords[actV].st[1]);
   actV=surfaces[i].triangles[j].indexes[2]; // odkaz na 3. vrchol trojuholnika
   surfaces[i].tris_t[ipos+1].set(surfaces[i].texcoords[actV].st[0],surfaces[i].texcoords[actV].st[1]);
   ipos+=3; // posun v poli texturovacich suradnic
  } // pre vsetky trojuholniky

  // normaly
  //au=(int) surfaces[k].vertices[ofsV+actV].n[0];
  //av=(int) surfaces[k].vertices[ofsV+actV].n[1];
  // dekodovanie normal
  //n[0]=nxyz[au][av][0];
  //n[1]=nxyz[au][av][1];
  //n[2]=nxyz[au][av][2];

  // uvolnenie vyhradenej pamate - tieto udaje uz dalej nebudu potrebne
  DELETEA(surfaces[i].shaders); DELETEA(surfaces[i].triangles);
  DELETEA(surfaces[i].texcoords); DELETEA(surfaces[i].vertices);
 } // pre vsetky sietove modely

 // reset prepojeni
 loopi(header.numTags) links[i]=NULL;

 // posledna snimka pre Q3 model je header.numFrames-1
 header.numFrames--;

 // zatvorenie MD3 suboru
 fclose(fp);

 // nahratie textury - iba pre 1. sietovy model
 char tmpName[255]; // pomocna premenna
 strcpy(tmpName,FNAME);
 ipos=strlen(tmpName)-1; while (tmpName[ipos]!='/') ipos--;
 // textura zbrane
 tmpName[ipos]='\0'; strcat(tmpName,"/skin.jpg\0");
 texture.load2D(tid[3],tmpName,T_MIPMAP|T_LINEAR);
 // textura dier po gulkach
 tmpName[ipos]='\0'; strcat(tmpName,"/holes_concrete.png\0");
 texture.load2D(tid[4],tmpName,T_MIPMAP|T_LINEAR|T_CLAMP);

 return(0);
}

// vykreslenie 1 konkretnej snimky *****************************************
void CLoadMD3::drawModel1F(int AF)
{
 glMatrixMode(GL_MODELVIEW); // modelview matica
 glPushMatrix();             // ulozenie
 glLoadIdentity();           // reset

  // textura zbrane
 glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[3]]);

 glEnableClientState(GL_VERTEX_ARRAY);
 glEnableClientState(GL_TEXTURE_COORD_ARRAY);

 loopk(header.numSurfaces) { // pre vsetky sietove modely
  glVertexPointer(3,GL_FLOAT,0,(surfaces[k].tris_v+3*AF*surfaces[k].header.numTriangles));
  glTexCoordPointer(2,GL_FLOAT,0,surfaces[k].tris_t);

  glDrawArrays(GL_TRIANGLES,0,3*surfaces[k].header.numTriangles);
 } // pre vsetky sietove modely

 glDisableClientState(GL_VERTEX_ARRAY);
 glDisableClientState(GL_TEXTURE_COORD_ARRAY);

 glPopMatrix();              // modelview matica
}

// vykreslenie MD3 modelu **************************************************
void CLoadMD3::draw(vec3 EP,vec3 TP,vec3 RP,GLubyte F)
{
 if ((F!=0)&&(stopaction==0)) { // ak je zbran v pokoji, avatar moze vystrelit
  stopaction=1; // zbran uz nie je v pokoji => avatar nemoze znovu vystrelit

  // pridanie odtlacku (decal) --------------------------------------------

  // normalizovane vektory smerujuce dopredu, doprava a hore
  frontV=TP; rightV=RP; upV=frontV.cross(rightV);

  ip0=wRange;      // maximalny dostrel
  TP=EP+TP*wRange; // bod na pozicii maximalneho dostrelu

  // pre vsetky trojuholniky
  loopvk(tris) {
   // ak luc zasiahol trojuholnik
   if (m3d.ray_tri_collision(-EP,-TP,tris[k].n,tris[k].d,tris[k].cn[0],tris[k].cn[1],tris[k].cn[2])) {
    pV=m3d.get_ipoint(); // priesecnik luca a trojuholnika
    ip1=(pV+EP).magnitude();
    if (ip1<ip0) { // vyber najblizsieho priesecnika
     ipoint=pV; inormal=tris[k].n; ip0=ip1;
    }
   }
  }

  // ak avatar zasiahol nejaky trojuholnik
  if (ip0!=wRange) {
   // priprava 6-tich orezavacich rovin (steny 3D hranola)
   cplane[0].x=rightV.x; cplane[0].y=rightV.y; cplane[0].z=rightV.z;    // prava stena
   pV=ipoint+rightV*wDecalHalfSize;
   cplane[0].w=-(cplane[0].x*pV.x+cplane[0].y*pV.y+cplane[0].z*pV.z);

   cplane[1].x=-rightV.x; cplane[1].y=-rightV.y; cplane[1].z=-rightV.z; // lava stena
   pV=ipoint-rightV*wDecalHalfSize;
   cplane[1].w=-(cplane[1].x*pV.x+cplane[1].y*pV.y+cplane[1].z*pV.z);

   cplane[2].x=upV.x; cplane[2].y=upV.y; cplane[2].z=upV.z;             // vrchna stena
   pV=ipoint+upV*wDecalHalfSize;
   cplane[2].w=-(cplane[2].x*pV.x+cplane[2].y*pV.y+cplane[2].z*pV.z);

   cplane[3].x=-upV.x; cplane[3].y=-upV.y; cplane[3].z=-upV.z;          // spodna stena
   pV=ipoint-upV*wDecalHalfSize;
   cplane[3].w=-(cplane[3].x*pV.x+cplane[3].y*pV.y+cplane[3].z*pV.z);

   cplane[4].x=frontV.x; cplane[4].y=frontV.y; cplane[4].z=frontV.z;    // zadna stena
   pV=ipoint+frontV*wDecalHalfSize;
   cplane[4].w=-(cplane[4].x*pV.x+cplane[4].y*pV.y+cplane[4].z*pV.z);

   cplane[5].x=-frontV.x; cplane[5].y=-frontV.y; cplane[5].z=-frontV.z; // predna stena
   pV=ipoint-frontV*wDecalHalfSize;
   cplane[5].w=-(cplane[5].x*pV.x+cplane[5].y*pV.y+cplane[5].z*pV.z);

   addDecal(cplane); // pridanie odtlacku (vsetkych casti odtlacku)
  } // ak avatar zasiahol nejaky trojuholnik
 }

 actframe+=max(1,(int) ceil(50.0f*difTime))*stopaction;
 if (actframe>header.numFrames) { actframe=0; stopaction=0; }
 drawModel1F(actframe); // vykreslenie 1 konkretnej snimky
}

// pridanie odtlacku (vsetkych casti odtlacku) *****************************
int CLoadMD3::addDecal(vec4 *P)
{
 // oznacenie casti posledneho pridaneho odtlacku
 lastG=NULL;

 // pre vsetky trojuholniky
 loopvk(tris) {
  // ak trojuholnik zasahuje do 3D hranola
  if (m3d.tri_in_box(tris[k].v[0],tris[k].v[1],tris[k].v[2],P)) {

   if (firstD==NULL) { // vyhradenie pamate pre 1. odtlacok
    firstD=new wDecal;
    if (firstD==NULL) {
     printf("Error allocating memory for weapon decal!\n"); return(1);
    }
    lastD=firstD;
   }
   else {              // vyhradenie pamate pre dalsi odtlacok
    lastD->next=new wDecal;
    if (lastD->next==NULL) {
     printf("Error allocating memory for weapon decal!\n"); return(1);
    }
    lastD=lastD->next;
   }

   // nasledujuci odtlacok
   lastD->next=NULL;

   // vrcholy casti odtlacku = vrcholy trojuholnika
   lastD->v[0]=tris[k].v[0]; lastD->v[1]=tris[k].v[1]; lastD->v[2]=tris[k].v[2];
   // normala = normala trojuholnika
   lastD->n=tris[k].n;
   // identifikator zavisly na normale, ak je normala casti odtlacku
   // postacujuco rovnobezna s inormal, potom idn=1 => tato cast
   // odtlacku lezi v rovine zasiahnutej gulkou
   d1=fabs(lastD->n.dot(inormal));
   if (d1>0.80f) lastD->idn=1; else lastD->idn=0;
   // orezanie casti odtlacku pomocou 6-tich stien 3D hranola
   lastD->vnum=m3d.tri_box_clip(lastD->v,P);

   // ak ma cast odtlacku po orezani aspon 3 vrcholy
   if (lastD->vnum>=3) {
    // ak este neboli oznacene casti posledneho pridaneho odtlacku
    if (lastG==NULL) lastG=lastD;
    // priesvitnost
    lastD->alpha=1.0f;
    // 8 vrcholov 3D hranola
    //lastD->abcd[0]=ipoint-rightV*wDecalHalfSize+frontV*wDecalHalfSize-upV*wDecalHalfSize;
    //lastD->abcd[1]=ipoint+rightV*wDecalHalfSize+frontV*wDecalHalfSize-upV*wDecalHalfSize;
    //lastD->abcd[2]=ipoint+rightV*wDecalHalfSize+frontV*wDecalHalfSize+upV*wDecalHalfSize;
    //lastD->abcd[3]=ipoint-rightV*wDecalHalfSize+frontV*wDecalHalfSize+upV*wDecalHalfSize;
    //lastD->abcd[4]=ipoint-rightV*wDecalHalfSize-frontV*wDecalHalfSize-upV*wDecalHalfSize;
    //lastD->abcd[5]=ipoint+rightV*wDecalHalfSize-frontV*wDecalHalfSize-upV*wDecalHalfSize;
    //lastD->abcd[6]=ipoint+rightV*wDecalHalfSize-frontV*wDecalHalfSize+upV*wDecalHalfSize;
    //lastD->abcd[7]=ipoint-rightV*wDecalHalfSize-frontV*wDecalHalfSize+upV*wDecalHalfSize;
   }
   else delDecal(lastD); // ak nema cast odtlacku min. 3 vrcholy, mozno ju vymazat

  } // ak trojuholnik zasahuje do 3D hranola
 } // pre vsetky trojuholniky

 // ak neboli oznacene casti posledneho pridaneho odtlacku => nebol pridany novy
 // odtlacok a teda nemusi prebehnut vypocet texturovacich suradnic
 if (lastG==NULL) return(1);

 // vypocet texturovacich suradnic -----------------------------------------

 // uprava velkosti odtlacku pre texturovanie
 ip0=1.0f/(wDecalHalfSize*4.0f); ip1=1.0f/(wDecalHalfSize*4.0f);
 // nahodny vyber 1 zo 4 textur, ktore reprezentuju odtlacok
 if (rand()<(RAND_MAX/2)) texU=0.25f; else texU=0.75f;
 if (rand()<(RAND_MAX/2)) texV=0.25f; else texV=0.75f;
 // vypocet uhlov pre otocenie vrcholov casti odtlacku do spolocnej roviny,
 // v ktorej prebehne vypocet texturovacich suradnic
 angR=PI2-inormal.angle(rightV); angU=PI2-inormal.angle(upV);

 // pre vsetky casti posledneho pridaneho odtlacku
 curD=lastG;
 while (curD!=NULL) {

  ind=0; // pociatocna hodnota indikatora (hrana, vrchol, neotacat)

  // ak normala casti odtlacku != inormala
  if (curD->idn==0) {
   // aktualna cast posledneho pridaneho odtlacku sa porovna s castami,
   // ktore lezia v rovine zasiahnutej gulkou
   cD=lastG;
   while (cD!=NULL) {
    // ak ide o rovnake casti, alebo ak porovnavana cast nelezi v rovine zasiahnutej gulkou
    if ((cD==curD)||(cD->idn==0)) { cD=cD->next; continue; }

    // najdenie spolocnej hrany 
    loopj(curD->vnum-1) { // pre vsetky hrany casti odtlacku
     loopi(cD->vnum-1) {  // pre vsetky hrany porovnavanej casti odtlacku
      // vypocet vzdialenosti vrcholov
      u1=(curD->v[j]-cD->v[i]);     // 1. vrchol - 1. vrchol
      u2=(curD->v[j+1]-cD->v[i+1]); // 2. vrchol - 2. vrchol
      u3=(curD->v[j]-cD->v[i+1]);   // 1. vrchol - 2. vrchol
      u4=(curD->v[j+1]-cD->v[i]);   // 2. vrchol - 1. vrchol
      d1=u1.magnitude(); d2=u2.magnitude(); d3=u3.magnitude(); d4=u4.magnitude();
      // ak su vrcholy hran postacujuco blizko, bola najdena spolocna hrana
      if (((d1<0.02f)&&(d2<0.02f))||((d3<0.02f)&&(d4<0.02f))) {
       u1=curD->v[j+1]-curD->v[j]; // hrana
       u2=curD->v[j]+u1*0.5f;      // vztazny bod
       u1.normalize();
       ind=1; break; // bola najdena spolocna hrana
      }
     }
     if (ind!=0) break; // ukoncenie cyklu loopj ak ind!=0
    }
    if (ind!=0) break; // ukoncenie cyklu while

    // najdenie spolocneho vrcholu
    loopj(curD->vnum) { // pre vsetky vrcholy casti odtlacku
     loopi(cD->vnum) {  // pre vsetky vrcholy porovnavanej casti odtlacku
      // vypocet vzdialenosti vrcholov
      u1=(curD->v[j]-cD->v[i]);
      d1=u1.magnitude();
      // ak su vrcholy postacujuco blizko, bol najdeny spolocny vrchol
      if (d1<0.02f) {
       // vrcholy hrany porovnavanej casti odtlacku
       u1=cD->v[i]; if ((i+1)<cD->vnum) u2=cD->v[i+1]; else u2=cD->v[0];
       u1=u1-u2;             // 1. vektor pre otocenie
       u1.normalize();
       u2=inormal.cross(u1); // 2. vektor pre otocenie
       u3=cD->v[i];          // vztazny bod
       ind=2; break; // bol najdeny spolocny vrchol
      }
     }
     if (ind!=0) break; // ukoncenie cyklu loopj ak ind!=0
    }
    if (ind!=0) break; // ukoncenie cyklu while

    cD=cD->next;
   }
   // ak nebola najdena spolocna hrana ani spolocny vrchol, cast odtlacku
   // sa nebude nijako otacat a priamo sa otexturuje, to sposobi vznik
   // nepeknych casti odtlackov, tuto nedokonalost vsak mozme zanedbat
   if (ind==0) ind=3;
  }

  if (ind==1) { // ak bola najdena spolocna hrana
   // vypocet uhla pre otocenie vrcholov okolo spolocnej hrany
   d3=inormal.angle(curD->n);
   // pre vsetky vrcholy casti odtlacku
   loopk(curD->vnum) {
    // posun vrcholu voci vztaznemu bodu
    pV=curD->v[k]-u2;
    // otacat sa bude v kladnom / zapornom smere
    d2=pV.x*inormal.x+pV.y*inormal.y+pV.z*inormal.z;
    if (d2>=0.0f) pV.rotate(-d3,u1); else pV.rotate(d3,u1);
    // posun vrcholu na povodnu poziciu a nasledny posun voci ipoint
    pV=(pV+u2)-ipoint;
    // otocenie vrcholu do spolocnej roviny
    pV.rotate(angU,rightV); pV.rotate(-angR,upV);
    // vypocet texturovacich suradnic
    d1=pV.dot(rightV)*ip0+texU; d2=pV.dot(upV)*ip1+texV;
    // orezanie texturovacich suradnic
    d1=max(texU-0.25f,min(d1,texU+0.25f)); d2=max(texV-0.25f,min(d2,texV+0.25f));
    // priradenie texturovacich suradnic konkretnemu vrcholu casti odtlacku
    curD->t[k].set(d1,d2);
    //curD->abcd[k]=ipoint+pV;
   }
  }
  else if (ind==2) { // ak bol najdeny spolocny vrchol
        // vypocet uhlov pre otocenie vrcholov okolo spolocneho vrcholu
        d3=PI2-curD->n.angle(u1); d4=PI2-curD->n.angle(u2);
        // pre vsetky vrcholy casti odtlacku
        loopk(curD->vnum) {
         // posun vrcholu voci vztaznemu bodu
         pV=curD->v[k]-u3;
         pV.rotate(d4,u1); pV.rotate(-d3,u2);
         // posun vrcholu na povodnu poziciu a nasledny posun voci ipoint
         pV=(pV+u3)-ipoint;
         // otocenie vrcholu do spolocnej roviny
         pV.rotate(angU,rightV); pV.rotate(-angR,upV);
         // vypocet texturovacich suradnic
         d1=pV.dot(rightV)*ip0+texU; d2=pV.dot(upV)*ip1+texV;
         // orezanie texturovacich suradnic
         d1=max(texU-0.25f,min(d1,texU+0.25f)); d2=max(texV-0.25f,min(d2,texV+0.25f));
         // priradenie texturovacich suradnic konkretnemu vrcholu casti odtlacku
         curD->t[k].set(d1,d2);
         //curD->abcd[k]=ipoint+pV;
        }
       }
       else {
        // pre vsetky vrcholy casti odtlacku
        loopk(curD->vnum) {
         // posun vrcholu voci ipoint
         pV=curD->v[k]-ipoint;
         // ak cast odtlacku lezi v rovine zasiahnutej gulkou, jej vrcholy
         // sa otocia do spolocnej roviny, ak cast odtlacku nelezi v rovine
         // zasiahnutej gulkou a zaroven nebola najdena spolocna hrana ani
         // vrchol, cast odtlacku sa otexturuje bez otacania
         if (ind==0) {
          // otocenie vrcholu
          pV.rotate(angU,rightV); pV.rotate(-angR,upV);
         }
         // vypocet texturovacich suradnic
         d1=pV.dot(rightV)*ip0+texU; d2=pV.dot(upV)*ip1+texV;
         // orezanie texturovacich suradnic
         d1=max(texU-0.25f,min(d1,texU+0.25f)); d2=max(texV-0.25f,min(d2,texV+0.25f));
         // priradenie texturovacich suradnic konkretnemu vrcholu casti odtlacku
         curD->t[k].set(d1,d2);
         //curD->abcd[k]=ipoint+pV;
        }
       }

  curD=curD->next;
 } // pre vsetky casti posledneho pridaneho odtlacku

 return(0);
}

// vykreslenie dier po gulkach *********************************************
void CLoadMD3::drawDecals(void)
{
 // vypnutie zapisu do DepthBuffera
 glDepthMask(GL_FALSE);
 // posun Z-hodnot
 glEnable(GL_POLYGON_OFFSET_FILL);
 // priesvitnost odtlacku
 glEnable(GL_BLEND);
 glBlendFunc(GL_DST_COLOR,GL_SRC_COLOR);
 // textura odtlacku
 glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[4]]);
 // shadery
 decal_vp.PRGenable(); decal_vp.PRGbind();
 decal_fp.PRGenable(); decal_fp.PRGbind();

 // pouziju sa polia atributov vrcholov:
 // 0 - suradnice vrcholu
 // 8 - texturovacie suradnice
 glEnableVertexAttribArray_(0);
 glEnableVertexAttribArray_(8);

 // pre vsetky odtlacky (pre vsetky casti odtlackov)
 curD=firstD;
 while (curD!=NULL) {
  glColor4f(0.5f,0.5f,0.5f,curD->alpha);            // farba
  glVertexAttribPointer_(0,3,GL_FLOAT,0,0,curD->v); // vrcholy
  glVertexAttribPointer_(8,2,GL_FLOAT,0,0,curD->t); // texturovacie suradnice
  glDrawArrays(GL_TRIANGLE_FAN,0,curD->vnum);
/*
  decal_fp.PRGdisable(); decal_vp.PRGdisable();

  glColor4f(1.0f,1.0f,0.0f,1.0f);
  glBegin(GL_LINE_LOOP);
   loopl(curD->vnum) glVertex3fv(curD->abcd[l].v);
  glEnd();

  decal_vp.PRGenable(); decal_vp.PRGbind();
  decal_fp.PRGenable(); decal_fp.PRGbind();
*/
  // priesvitnost odtlacku
  curD->alpha-=0.004f; if (curD->alpha<0.0f) delDecal(curD);

  if (curD!=NULL) curD=curD->next;
 }

 // standardna farba
 glColor4f(1.0f,1.0f,1.0f,1.0f);

 glDisableVertexAttribArray_(0);
 glDisableVertexAttribArray_(8);

 // vypnutie shaderov
 decal_fp.PRGdisable(); decal_vp.PRGdisable();
 // vypnutie priesvitnosti
 glDisable(GL_BLEND);
 // vypnutie posunu Z-hodnot
 glDisable(GL_POLYGON_OFFSET_FILL);
 // zapnutie zapisu do DepthBuffera
 glDepthMask(GL_TRUE);
}

// vymazanie odtlacku, alebo casti odtlacku ********************************
void CLoadMD3::delDecal(wDecal *P)
{
 if (P==firstD) { // zmazanie 1. odtlacku
  firstD=firstD->next;
  DELETEP(P);
  curD=firstD;
 }
 else {           // zmazanie odtlacku v zozname
  curD=firstD;
  while (curD->next!=P) curD=curD->next;
  curD->next=P->next;
  DELETEP(P);
  if (curD->next==NULL) lastD=curD;
 }
}

// zmazanie vsetkych odtlackov (vsetkych casti) ****************************
void CLoadMD3::delAllDecals(void)
{
 curD=firstD;
 if (curD!=NULL) {
  while (curD->next!=NULL) { // zmazanie vsetkych odtlackov zo zoznamu
   firstD=curD->next;
   DELETEP(curD);
   curD=firstD;
  }
  DELETEP(firstD);           // zmazanie 1. odtlacku
 }

 // prvy, posledny, aktualny odtlacok (casti odtlacku),
 // oznacenie casti posledneho pridaneho odtlacku
 firstD=NULL; lastD=NULL; curD=NULL; cD=NULL; lastG=NULL;
}
