// hlavickove subory *******************************************************
#include "../includes.h"
// vlastny .h subor
#include "cloth.h"

// konstruktor *************************************************************
CCloth::CCloth()
{
 width=3.0f; height=4.0f;     // sirka/vyska platna
 width_num=30; height_num=40; // pocet castic v smere sirky/vysky platna

 particles.reset(); // zoznam castic
 particles_num=0;   // pocet castic

 cloth_tris.reset(); // zoznam trojuholnikov
}

// destruktor **************************************************************
CCloth::~CCloth()
{

}

// inicializacia platna ****************************************************
int CCloth::init(void)
{
 // priprava zoznamu castic
 loopj(height_num) {
  loopi(width_num) {
   // pridanie novej castice do zoznamu
   particles.add();
   // poloha castice
   pv.set(0.0f,-height*((float) j/(float) height_num),-width*((float) i/(float) width_num));
   pv+=vec3(-9.1f,3.0f,9.4f);
   particles[particles_num].pos=pv;
   // normala
   particles[particles_num].n.set(1.0f,0.0f,0.0f);
   // texturovacie suradnice
   dv.set(((float) i/(float) width_num),((float) j/(float) height_num));
   particles[particles_num].uv=dv;

   particles_num++; // zvysenie poctu castic
  }
 }

 // priprava zoznamu trojuholnikov
 loopj(height_num-1) {
  loopi(width_num-1) {
   // pridanie dvoch trojuholnikov
   cloth_tris.add(); cloth_tris.add();
  }
 }

 // priprava obmedzeni pohybu castic
 loopvk(particles) {
  // indexy do zoznamu castic
  p0=k%width_num; p1=k/width_num;
  // horizontalna vazba
  if (p0<width_num-1) {
   pv=particles[p0+1+p1*width_num].pos-particles[k].pos;
   cur_dist=pv.magnitude(); // vzdialenost p1 - p2
   particles[k].constraints[0].p=&particles[p0+1+p1*width_num];
   particles[k].constraints[0].rest_dist=cur_dist;
  }
  else particles[k].constraints[0].p=NULL;
  // diagonalna vazba - hlavna diagonala
  if ((p0<width_num-1)&&(p1<height_num-1)) {
   pv=particles[p0+1+(p1+1)*width_num].pos-particles[k].pos;
   cur_dist=pv.magnitude(); // vzdialenost p1 - p2
   particles[k].constraints[1].p=&particles[p0+1+(p1+1)*width_num];
   particles[k].constraints[1].rest_dist=cur_dist;
  }
  else particles[k].constraints[1].p=NULL;
  // vertikalna vazba
  if (p1<height_num-1) {
   pv=particles[p0+(p1+1)*width_num].pos-particles[k].pos;
   cur_dist=pv.magnitude(); // vzdialenost p1 - p2
   particles[k].constraints[2].p=&particles[p0+(p1+1)*width_num];
   particles[k].constraints[2].rest_dist=cur_dist;
  }
  else particles[k].constraints[2].p=NULL;
  // diagonalna vazba - vedlajsia diagonala
  if ((p0>0)&&(p1<height_num-1)) {
   pv=particles[p0-1+(p1+1)*width_num].pos-particles[k].pos;
   cur_dist=pv.magnitude(); // vzdialenost p1 - p2
   particles[k].constraints[3].p=&particles[p0-1+(p1+1)*width_num];
   particles[k].constraints[3].rest_dist=cur_dist;
  }
  else particles[k].constraints[3].p=NULL;
 }

 // priprava 6-tich nepohyblivych castic
 loopi(3) {
  particles[i].pos.z-=0.2f;               // posun castice pre realnejsi vzhlad platna
  particles[width_num-1-i].pos.z+=0.2f;   // posun castice pre realnejsi vzhlad platna
  particles[i].can_move=0.0f;             // castica sa nebude pohybovat
  particles[width_num-1-i].can_move=0.0f; // castica sa nebude pohybovat
 }

 // vektor posobenia sily gravitacie - gravitacne zrychlenie v danom smere
 Gv.set(0.0f,-9.81f,0.0f);
 // vektor posobenia sily vetra
 Wv.set(0.1f,0.0f,0.8f); Wv.normalize();

 // pomocny kolizny vektor
 Cv.set(10.0f,0.0f,0.0f);

 // nahratie textury platna
 texture.load2D(tid[100],"data/textures/sk.png",T_MIPMAP|T_LINEAR|T_CLAMP);

 return(0);
}

// simulacia fyzikalneho deja **********************************************
void CCloth::simulate(float DT,int CID)
{
 // pocet opakovani simulacie
 iterations_num=(int) (DT/MAXDT)+1;
 dt=DT/(float) iterations_num;
 // printf("%i\n",iterations_num);
 // ak bude DT prilis velke, simulacia sa zopakuje viackrat
 loopi(iterations_num) {
  // reset sil posobiacich na castice
  loopvk(particles) particles[k].force.set(0.0f,0.0f,0.0f);

  // pre vsetky castice
  loopvk(particles) {
   // indexy do zoznamu castic
   p0=k%width_num; p1=k/width_num;

   // pridanie pritazlivych sil, ktore posobia na zviazane castice
   loopj(4) {
    // ak konkretna vazba neexistuje, pokracuje sa na dalsej moznej vazbe
    if (particles[k].constraints[j].p==NULL) continue;

    pv=particles[k].constraints[j].p->pos-particles[k].pos; // vektor p1->p2
    cur_dist=pv.magnitude(); // velkost vektora
    pv/=cur_dist;            // normalizacia vektora
    // posobenie pritazlivej sily
    pv*=(cur_dist-particles[k].constraints[j].rest_dist)*SPRING_STIFFNESS;
    pv-=(particles[k].vel-particles[k].constraints[j].p->vel)*SPRING_FRICTION;
    // pridanie sil posobiacich na castice
    particles[k].addForce(pv);
    particles[k].constraints[j].p->addForce(-pv);
   }

   // posobenie sily gravitacie F=m*g
   particles[k].addForce(Gv*PMASS);

   // posobenie sily vetra
   if ((p0<width_num-1)&&(p1<height_num-1)) {
    vA=particles[p0+(p1+1)*width_num].pos-particles[k].pos;
    vB=particles[p0+1+(p1+1)*width_num].pos-particles[k].pos;
    vA=vA.cross(vB);
    vB=vA; vB.normalize();
    f=vB.dot(Wv);
    pv=vA*f;
    particles[k].addForce(pv);
    particles[p0+(p1+1)*width_num].addForce(pv);
    particles[p0+1+(p1+1)*width_num].addForce(pv);

    vA=particles[p0+1+(p1+1)*width_num].pos-particles[k].pos;
    vB=particles[p0+1+p1*width_num].pos-particles[k].pos;
    vA=vA.cross(vB);
    vB=vA; vB.normalize();
    f=vB.dot(Wv);
    pv=vA*f;
    particles[k].addForce(pv);
    particles[p0+1+(p1+1)*width_num].addForce(pv);
    particles[p0+1+p1*width_num].addForce(pv);
   }

   // trenie sposobene vzduchom
   particles[k].addForce(-particles[k].vel*0.1f);

   // na casticu, ktora kolidovala s gulkou posobi sila
   if (k==CID) { particles[k].addForce(Fv*PMASS); }

   // obmedzenie pohybu castice po narazeni do steny a podlahy - pre tento konkretny pripad
   if (particles[k].pos.x<=-9.9f) particles[k].addForce(Cv);
   if (particles[k].pos.y<=-1.3f) particles[k].addForce(-Gv);

   // simulacia fyzikalneho deja pre casticu
   particles[k].simulate(dt);
  }
 }
}

// vykreslenie platna ******************************************************
void CCloth::draw(vec3 EP,vec3 TP,GLubyte F)
{
 col_id=-1;

 // ak Avatar striela
 if (F!=0) {
  TP=EP+TP*wRange; // bod na pozicii maximalneho dostrelu
  // pre vsetky castice
  loopvk(particles) {
   // vypocet vzdialenosti castice od luca
   vA=particles[k].pos;
   vB=m3d.point_on_line(-EP,-TP,vA);
   pv=vB-vA; f=pv.magnitude();
   if (f<0.05f) {
    col_id=k; Fv=(-TP+EP)*50.0f; // vektor posobenia sily gulky
    // indexy do zoznamu castic
    p0=k%width_num; p1=k/width_num;
    if (p1>0) {
     // narusenie vzajomnych vazieb medzi casticami
     if (p0>0) particles[p0-1+(p1-1)*width_num].constraints[1].p=NULL;
     particles[p0+(p1-1)*width_num].constraints[2].p=NULL;
     if (p0<width_num-1) particles[p0+1+(p1-1)*width_num].constraints[3].p=NULL;
    }
    break;
   }
  }
 }

 // simulacia fyzikalneho deja pre platno
 simulate(difTime,col_id);

 // vypocet normal
 loopj(height_num-1) {
  loopi(width_num-1) {
   q=i+j*width_num;
   // 1. trojuholnik
   vA=particles[i+(j+1)*width_num].pos-particles[q].pos;
   vB=particles[i+1+(j+1)*width_num].pos-particles[q].pos;
   v0=vA.cross(vB);
   v0.normalize();
   // 2. trojuholnik
   vA=particles[i+1+(j+1)*width_num].pos-particles[q].pos;
   vB=particles[i+1+j*width_num].pos-particles[q].pos;
   v1=vA.cross(vB);
   v1.normalize();
   // normala
   pv=v0+v1; pv/=2.0f;
   particles[q].n=pv;
  }
 }
 // vyhladenie normal
 loopj(height_num-2) {
  if (j==0) continue;
  loopi(width_num-2) {
   if (i==0) continue;
   // normala=priemer 4 susediacich normal
   pv=particles[i-1+j*width_num].n+particles[i+(j+1)*width_num].n+
      particles[i+1+j*width_num].n+particles[i+(j-1)*width_num].n;
   pv/=4.0f;
   particles[i+j*width_num].n=pv;
  }
 }

 q=0;
 // naplnenie zoznamu trojuholnikov
 loopj(height_num-1) {
  loopi(width_num-1) {
   if ((particles[i+j*width_num].constraints[1].p!=NULL)&&
       (particles[i+j*width_num].constraints[2].p!=NULL)) {
    cloth_tris[q].v[0]=particles[i+j*width_num].pos;
    cloth_tris[q].v[1]=particles[i+(j+1)*width_num].pos;
    cloth_tris[q].v[2]=particles[i+1+(j+1)*width_num].pos;
    cloth_tris[q].vn[0]=particles[i+j*width_num].n;
    cloth_tris[q].vn[1]=particles[i+(j+1)*width_num].n;
    cloth_tris[q].vn[2]=particles[i+1+(j+1)*width_num].n;
    cloth_tris[q].t[0]=particles[i+j*width_num].uv;
    cloth_tris[q].t[1]=particles[i+(j+1)*width_num].uv;
    cloth_tris[q].t[2]=particles[i+1+(j+1)*width_num].uv;
    q++;
   }
   if ((particles[i+j*width_num].constraints[0].p!=NULL)&&
       (particles[i+j*width_num].constraints[1].p!=NULL)) {
    cloth_tris[q].v[0]=particles[i+j*width_num].pos;
    cloth_tris[q].v[1]=particles[i+1+(j+1)*width_num].pos;
    cloth_tris[q].v[2]=particles[i+1+j*width_num].pos;
    cloth_tris[q].vn[0]=particles[i+j*width_num].n;
    cloth_tris[q].vn[1]=particles[i+1+(j+1)*width_num].n;
    cloth_tris[q].vn[2]=particles[i+1+j*width_num].n;
    cloth_tris[q].t[0]=particles[i+j*width_num].uv;
    cloth_tris[q].t[1]=particles[i+1+(j+1)*width_num].uv;
    cloth_tris[q].t[2]=particles[i+1+j*width_num].uv;
    q++;
   }
  }
 }

 // vyber textury
 glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[100]]);

 // spracuju sa aj odvratene strany trojuholnikov
 glDisable(GL_CULL_FACE);
   
 if (glextGLSL) { // s GLSL shadermi
  glsl_shader.GLSLon(1);

  // pouziju sa polia atributov vrcholov:
  // 0 - suradnice vrcholu
  // 2 - normala
  // 8 - texturovacie suradnice
  glEnableVertexAttribArray_(0);
  glEnableVertexAttribArray_(2);
  glEnableVertexAttribArray_(8);

  // vykreslenie vsetkych trojuholnikov
  loopk(q) {
   glVertexAttribPointer_(0,3,GL_FLOAT,0,0,cloth_tris[k].v);  // vrchol
   glVertexAttribPointer_(2,3,GL_FLOAT,0,0,cloth_tris[k].vn); // normala
   glVertexAttribPointer_(8,2,GL_FLOAT,0,0,cloth_tris[k].t);  // texturovacie suradnice

   glDrawArrays(GL_TRIANGLES,0,3);
  }

  glDisableVertexAttribArray_(0);
  glDisableVertexAttribArray_(2);
  glDisableVertexAttribArray_(8);
 
  glsl_shader.GLSLoff();
 }
 else { // bez GLSL shaderov
  glEnable(GL_LIGHTING);
  glBegin(GL_TRIANGLES);
  loopk(q) {
   glTexCoord2fv(cloth_tris[k].t[0].v); glNormal3fv(cloth_tris[k].vn[0].v);
   glVertex3fv(cloth_tris[k].v[0].v);
   glTexCoord2fv(cloth_tris[k].t[1].v); glNormal3fv(cloth_tris[k].vn[1].v);
   glVertex3fv(cloth_tris[k].v[1].v);
   glTexCoord2fv(cloth_tris[k].t[2].v); glNormal3fv(cloth_tris[k].vn[2].v);
   glVertex3fv(cloth_tris[k].v[2].v);
  }
  glEnd();
  glDisable(GL_LIGHTING);
 }

 // odvratene strany trojuholnikov sa nebudu dalej spracovavat
 glEnable(GL_CULL_FACE);
}
